from keras.losses import binary_crossentropy
import keras.backend as K


def diceCoeff(yTrue, yPred):
    smooth = 1.
    yTrueF = K.flatten(yTrue)
    yPredF = K.flatten(yPred)
    intersection = K.sum(yTrueF * yPredF)
    score = (2. * intersection + smooth) / (K.sum(yTrueF) + K.sum(yPredF) + smooth)
    return score


def diceLoss(yTrue, yPred):
    loss = 1 - diceCoeff(yTrue, yPred)
    return loss


def bceDiceLoss(yTrue, yPred):
    loss = binary_crossentropy(yTrue, yPred) + diceLoss(yTrue, yPred)
    return loss


def weightedDiceCoeff(yTrue, yPred, weight):
    smooth = 1.
    w, m1, m2 = weight * weight, yTrue, yPred
    intersection = (m1 * m2)
    score = (2. * K.sum(w * intersection) + smooth) / (K.sum(w * m1) + K.sum(w * m2) + smooth)
    return score


def weightedDiceLoss(yTrue, yPred):
    yTrue = K.cast(yTrue, 'float32')
    yPred = K.cast(yPred, 'float32')
    # if we want to get same size of output, kernel size must be odd number
    if K.int_shape(yPred)[1] == 128:
        kernelSize = 11
    elif K.int_shape(yPred)[1] == 256:
        kernelSize = 21
    elif K.int_shape(yPred)[1] == 512:
        kernelSize = 21
    elif K.int_shape(yPred)[1] == 1024:
        kernelSize = 41
    else:
        raise ValueError('Unexpected image size')
    averagedMask = K.pool2d(
        yTrue, pool_size=(kernelSize, kernelSize), strides=(1, 1), padding='same', pool_mode='avg')
    border = K.cast(K.greater(averagedMask, 0.005), 'float32') * K.cast(K.less(averagedMask, 0.995), 'float32')
    weight = K.ones_like(averagedMask)
    w0 = K.sum(weight)
    weight += border * 2
    w1 = K.sum(weight)
    weight *= (w0 / w1)
    loss = 1 - weightedDiceCoeff(yTrue, yPred, weight)
    return loss


def weightedBceLoss(yTrue, yPred, weight):
    # avoiding overflow
    epsilon = 1e-7
    yPred = K.clip(yPred, epsilon, 1. - epsilon)
    logitYPred = K.log(yPred / (1. - yPred))

    # https://www.tensorflow.org/api_docs/python/tf/nn/weighted_cross_entropy_with_logits
    loss = (1. - yTrue) * logitYPred + (1. + (weight - 1.) * yTrue) * \
                                          (K.log(1. + K.exp(-K.abs(logitYPred))) + K.maximum(-logitYPred, 0.))
    return K.sum(loss) / K.sum(weight)


def weightedBceDiceLoss(y_true, yPred):
    yTrue = K.cast(yTrue, 'float32')
    yPred = K.cast(yPred, 'float32')
    # if we want to get same size of output, kernel size must be odd number
    if K.int_shape(yPred)[1] == 128:
        kernelSize = 11
    elif K.int_shape(yPred)[1] == 256:
        kernelSize = 21
    elif K.int_shape(yPred)[1] == 512:
        kernelSize = 21
    elif K.int_shape(yPred)[1] == 1024:
        kernelSize = 41
    else:
        raise ValueError('Unexpected image size')
    averagedMask = K.pool2d(
        yTrue, pool_size=(kernelSize, kernelSize), strides=(1, 1), padding='same', pool_mode='avg')
    border = K.cast(K.greater(averagedMask, 0.005), 'float32') * K.cast(K.less(averagedMask, 0.995), 'float32')
    weight = K.ones_like(averagedMask)
    w0 = K.sum(weight)
    weight += border * 2
    w1 = K.sum(weight)
    weight *= (w0 / w1)
    loss = weightedBceLoss(yTrue, yPred, weight) + (1 - weightedDiceCoeff(yTrue, yPred, weight))
    return loss
