from keras.models import Model
from keras.layers import *
from keras.optimizers import *
from keras import backend as K
from model.losses import bce_dice_loss, dice_loss, weighted_bce_dice_loss, weighted_dice_loss, dice_coeff
from keras.losses import binary_crossentropy 

def dice_coeff(y_true, y_pred):
    smooth = 1.
    y_true_f = K.flatten(y_true)
    y_pred_f = K.flatten(y_pred)
    intersection = K.sum(y_true_f * y_pred_f)
    score = (2. * intersection + smooth) / (K.sum(y_true_f) + K.sum(y_pred_f) + smooth)
    return score
    
def dice_loss(y_true, y_pred):
    loss = 1 - dice_coeff(y_true, y_pred)
    return loss


def bce_dice_loss(y_true, y_pred):
    loss = binary_crossentropy(y_true, y_pred) + dice_loss(y_true, y_pred)
    return loss


def unet_base(input_shape=(3,128, 128),
                 num_classes=1,learning_r=0.001, batch_size = 32):
    inputs = Input(shape=input_shape)

    in_normal = BatchNormalization()(inputs)
    n_filters = 24
    
    # Contracting step 1
    d1 = Conv2D(n_filters, (3, 3),  kernel_initializer='he_normal', padding='same') (in_normal)
    d1 = ReLU()(d1)
    d1 = Dropout(0.1)(d1)
    d1 = Conv2D(n_filters, (3, 3),  kernel_initializer='he_normal', padding='same') (d1)
    d1 = ReLU()(d1)
    p1 = MaxPooling2D((2, 2), strides=(2, 2))(d1)
   
    # Contracting step 2
    d2 = Conv2D(2*n_filters, (3, 3),  kernel_initializer='he_normal', padding='same') (p1)
    d2 = ReLU()(d2)
    d2 = Dropout(0.2)(d2)
    d2 = Conv2D(2*n_filters, (3, 3),  kernel_initializer='he_normal', padding='same') (d2)
    d2 = ReLU()(d2)
    p2 = MaxPooling2D((2, 2), strides=(2, 2))(d2)
    
    # Contracting step 3
    d3 = Conv2D(4*n_filters, (3, 3),  kernel_initializer='he_normal', padding='same') (p2)
    d3 = ReLU()(d3)
    d3 = Dropout(0.2)(d3)
    d3 = Conv2D(4*n_filters, (3, 3),  kernel_initializer='he_normal', padding='same') (d3)
    d3 = ReLU()(d3)
    p3 = MaxPooling2D((2, 2), strides=(2, 2))(d3)
    
    # Contracting step 4
    d4 = Conv2D(8*n_filters, (3, 3),  kernel_initializer='he_normal', padding='same') (p3)
    d4 = ReLU()(d4)
    d4 = Dropout(0.2)(d4)
    d4 = Conv2D(8*n_filters, (3, 3),  kernel_initializer='he_normal', padding='same') (d4)
    d4 = ReLU()(d4)
    p4 = MaxPooling2D((2, 2), strides=(2, 2))(d4)
    
    # Center
    c = Conv2D(16*n_filters, (3, 3),  kernel_initializer='he_normal', padding='same') (p4)
    c = ReLU()(c)
    c = Dropout(0.3)(c)
    c = Conv2D(16*n_filters, (3, 3),  kernel_initializer='he_normal', padding='same') (c)
    c = ReLU()(c)
    
    # Expanding step 1
    u4 = UpSampling2D((2,2))(c)
    u4 = Conv2D(8*n_filters, (2, 2),  kernel_initializer='he_normal', padding='same') (u4)
    u4 = concatenate([d4, u4], axis=1)
    u4 = Conv2D(8*n_filters, (3, 3),  kernel_initializer='he_normal', padding='same') (u4)
    u4 = ReLU()(u4)
    u4 = Dropout(0.2)(u4)
    u4 = Conv2D(8*n_filters, (3, 3),  kernel_initializer='he_normal', padding='same') (u4)
    u4 = ReLU()(u4)


    # Expanding step 2
    u3 = UpSampling2D((2,2))(u4)
    u3 = Conv2D(4*n_filters, (2, 2),  kernel_initializer='he_normal', padding='same') (u3)
    u3 = concatenate([d3, u3], axis=1)
    u3 = Conv2D(4*n_filters, (3, 3),  kernel_initializer='he_normal', padding='same') (u3)
    u3 = ReLU()(u3)
    u3 = Dropout(0.2)(u3)
    u3 = Conv2D(4*n_filters, (3, 3),  kernel_initializer='he_normal', padding='same') (u3)
    u3 = ReLU()(u3)
    
    # Expanding step 3
    u2 = UpSampling2D((2,2))(u3)
    u2 = Conv2D(2*n_filters, (2, 2),  kernel_initializer='he_normal', padding='same') (u2)
    u2 = concatenate([d2, u2], axis=1)
    u2 = Conv2D(2*n_filters, (3, 3),  kernel_initializer='he_normal', padding='same') (u2)
    u2 = ReLU()(u2)
    u2 = Dropout(0.2)(u2)
    u2 = Conv2D(2*n_filters, (3, 3),  kernel_initializer='he_normal', padding='same') (u2)
    u2 = ReLU()(u2)
    
    # Expanding step 4
    u1 = UpSampling2D((2,2))(u2)
    u1 = Conv2D(n_filters, (2, 2),  kernel_initializer='he_normal', padding='same') (u1)
    u1 = concatenate([d1, u1], axis=1)
    u1 = Conv2D(n_filters, (3, 3),  kernel_initializer='he_normal', padding='same') (u1)
    u1 = ReLU()(u1)
    u1 = Dropout(0.1)(u1)
    u1 = Conv2D(n_filters, (3, 3),  kernel_initializer='he_normal', padding='same') (u1)
    u1 = ReLU()(u1)
    
    # Final layer
    classify = Conv2D(num_classes, (1, 1), activation='sigmoid')(u1)
    model = Model(inputs=inputs, outputs=classify)

    model.compile(optimizer=Nadam(rescale_grad=1/batch_size,lr=learning_r,beta_1=0.95, beta_2=0.99), loss=bce_dice_loss, metrics=[dice_coeff])
    return model
